# Copyright 2010-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from itertools import chain
import stat
import subprocess

from portage.const import BASH_BINARY, PORTAGE_BASE_PATH, PORTAGE_BIN_PATH
from portage.tests import TestCase
from portage import os
from portage import _encodings
from portage import _unicode_decode, _unicode_encode


class BashSyntaxTestCase(TestCase):
    def testBashSyntax(self):
        locations = [PORTAGE_BIN_PATH]
        misc_dir = os.path.join(PORTAGE_BASE_PATH, "misc")
        if os.path.isdir(misc_dir):
            locations.append(misc_dir)
        for parent, dirs, files in chain.from_iterable(os.walk(x) for x in locations):
            parent = _unicode_decode(parent, encoding=_encodings["fs"], errors="strict")
            for x in files:
                x = _unicode_decode(x, encoding=_encodings["fs"], errors="strict")
                ext = x.split(".")[-1]
                if ext in (".py", ".pyc", ".pyo"):
                    continue
                x = os.path.join(parent, x)
                st = os.lstat(x)
                if not stat.S_ISREG(st.st_mode):
                    continue

                # Check for bash shebang
                f = open(
                    _unicode_encode(x, encoding=_encodings["fs"], errors="strict"), "rb"
                )
                line = _unicode_decode(
                    f.readline(), encoding=_encodings["content"], errors="replace"
                )
                f.close()
                if line[:2] == "#!" and "bash" in line:
                    cmd = [BASH_BINARY, "-n", x]
                    cmd = [
                        _unicode_encode(x, encoding=_encodings["fs"], errors="strict")
                        for x in cmd
                    ]
                    proc = subprocess.Popen(
                        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                    )
                    output = _unicode_decode(
                        proc.communicate()[0], encoding=_encodings["fs"]
                    )
                    status = proc.wait()
                    self.assertEqual(
                        os.WIFEXITED(status) and os.WEXITSTATUS(status) == os.EX_OK,
                        True,
                        msg=output,
                    )



# Contains code from https://github.com/MagicStack/uvloop/tree/v0.16.0
# SPDX-License-Identifier: PSF-2.0 AND (MIT OR Apache-2.0)
# SPDX-FileCopyrightText: Copyright (c) 2015-2021 MagicStack Inc.  http://magic.io

import enum

# After the connection is lost, log warnings after this many write()s.
LOG_THRESHOLD_FOR_CONNLOST_WRITES = 5

# Seconds to wait before retrying accept().
ACCEPT_RETRY_DELAY = 1

# Number of stack entries to capture in debug mode.
# The larger the number, the slower the operation in debug mode
# (see extract_stack() in format_helpers.py).
DEBUG_STACK_DEPTH = 10

# Number of seconds to wait for SSL handshake to complete
# The default timeout matches that of Nginx.
SSL_HANDSHAKE_TIMEOUT = 60.0

# Number of seconds to wait for SSL shutdown to complete
# The default timeout mimics lingering_time
SSL_SHUTDOWN_TIMEOUT = 30.0

# Used in sendfile fallback code.  We use fallback for platforms
# that don't support sendfile, or for TLS connections.
SENDFILE_FALLBACK_READBUFFER_SIZE = 1024 * 256

FLOW_CONTROL_HIGH_WATER_SSL_READ = 256  # KiB
FLOW_CONTROL_HIGH_WATER_SSL_WRITE = 512  # KiB

# Default timeout for joining the threads in the threadpool
THREAD_JOIN_TIMEOUT = 300

# The enum should be here to break circular dependencies between
# base_events and sslproto
class _SendfileMode(enum.Enum):
    UNSUPPORTED = enum.auto()
    TRY_NATIVE = enum.auto()
    FALLBACK = enum.auto()



# Copyright: 2005-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# Author(s): Brian Harring (ferringb@gentoo.org)

__all__ = [
    "Mapping",
    "MutableMapping",
    "UserDict",
    "ProtectedDict",
    "LazyLoad",
    "slot_dict_class",
]

import weakref


class Mapping:
    """
    In python-3.0, the UserDict.DictMixin class has been replaced by
    Mapping and MutableMapping from the collections module, but 2to3
    doesn't currently account for this change:

        https://bugs.python.org/issue2876

    As a workaround for the above issue, use this class as a substitute
    for UserDict.DictMixin so that code converted via 2to3 will run.
    """

    __slots__ = ()

    def __iter__(self):
        return iter(self.keys())

    def __contains__(self, key):
        try:
            value = self[key]
        except KeyError:
            return False
        return True

    def iteritems(self):
        for k in self:
            yield (k, self[k])

    def iterkeys(self):
        return self.__iter__()

    def itervalues(self):
        for _, v in self.items():
            yield v

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __repr__(self):
        return repr(dict(self.items()))

    def __len__(self):
        return len(list(self))

    # TODO: do we need to keep iter*?
    items = iteritems
    keys = __iter__
    values = itervalues


class MutableMapping(Mapping):
    """
    A mutable version of the Mapping class.
    """

    __slots__ = ()

    def clear(self):
        for key in list(self):
            del self[key]

    def setdefault(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return default

    def pop(self, key, *args):
        if len(args) > 1:
            raise TypeError(
                "pop expected at most 2 arguments, got " + repr(1 + len(args))
            )
        try:
            value = self[key]
        except KeyError:
            if args:
                return args[0]
            raise
        del self[key]
        return value

    def popitem(self):
        try:
            k, v = next(iter(self.items()))
        except StopIteration:
            raise KeyError("container is empty")
        del self[k]
        return (k, v)

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError(
                "expected at most 1 positional argument, got " + repr(len(args))
            )
        other = None
        if args:
            other = args[0]
        if other is None:
            pass
        elif hasattr(other, "iteritems"):
            # Use getattr to avoid interference from 2to3.
            for k, v in getattr(other, "iteritems")():
                self[k] = v
        elif hasattr(other, "items"):
            # Use getattr to avoid interference from 2to3.
            for k, v in getattr(other, "items")():
                self[k] = v
        elif hasattr(other, "keys"):
            for k in other.keys():
                self[k] = other[k]
        else:
            for k, v in other:
                self[k] = v
        if kwargs:
            self.update(kwargs)


class UserDict(MutableMapping):
    """
    Use this class as a substitute for UserDict.UserDict so that
    code converted via 2to3 will run:

         https://bugs.python.org/issue2876
    """

    __slots__ = ("data",)

    def __init__(self, *args, **kwargs):
        self.data = {}

        if len(args) > 1:
            raise TypeError(
                "expected at most 1 positional argument, got " + repr(len(args))
            )

        if args:
            self.update(args[0])

        if kwargs:
            self.update(kwargs)

    def __repr__(self):
        return repr(self.data)

    def __contains__(self, key):
        return key in self.data

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, item):
        self.data[key] = item

    def __delitem__(self, key):
        del self.data[key]

    def clear(self):
        self.data.clear()

    keys = __iter__


class ProtectedDict(MutableMapping):
    """
    given an initial dict, this wraps that dict storing changes in a secondary dict, protecting
    the underlying dict from changes
    """

    __slots__ = ("orig", "new", "blacklist")

    def __init__(self, orig):
        self.orig = orig
        self.new = {}
        self.blacklist = {}

    def __setitem__(self, key, val):
        self.new[key] = val
        if key in self.blacklist:
            del self.blacklist[key]

    def __getitem__(self, key):
        if key in self.new:
            return self.new[key]
        if key in self.blacklist:
            raise KeyError(key)
        return self.orig[key]

    def __delitem__(self, key):
        if key in self.new:
            del self.new[key]
        elif key in self.orig:
            if key not in self.blacklist:
                self.blacklist[key] = True
                return
        raise KeyError(key)

    def __iter__(self):
        for k in self.new:
            yield k
        for k in self.orig:
            if k not in self.blacklist and k not in self.new:
                yield k

    def __contains__(self, key):
        return key in self.new or (key not in self.blacklist and key in self.orig)

    keys = __iter__


class LazyLoad(Mapping):
    """
    Lazy loading of values for a dict
    """

    __slots__ = ("pull", "d")

    def __init__(self, pull_items_func, initial_items=[]):
        self.d = {}
        for k, v in initial_items:
            self.d[k] = v
        self.pull = pull_items_func

    def __getitem__(self, key):
        if key in self.d:
            return self.d[key]
        if self.pull is not None:
            self.d.update(self.pull())
            self.pull = None
        return self.d[key]

    def __iter__(self):
        if self.pull is not None:
            self.d.update(self.pull())
            self.pull = None
        return iter(self.d)

    def __contains__(self, key):
        if key in self.d:
            return True
        if self.pull is not None:
            self.d.update(self.pull())
            self.pull = None
        return key in self.d

    keys = __iter__


class _SlotDict:
    """
    Base class for classes returned from slot_dict_class.
    """

    _prefix = ""
    allowed_keys = frozenset()
    __slots__ = ("__weakref__",)

    def __init__(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError(
                "expected at most 1 positional argument, got " + repr(len(args))
            )

        if args:
            self.update(args[0])

        if kwargs:
            self.update(kwargs)

    def __reduce__(self):
        return _PickledSlotDict, (
            self._prefix,
            self.allowed_keys,
            dict(self),
        )

    def __eq__(self, other):
        return dict(self) == dict(other)

    def __iter__(self):
        for k, v in self.iteritems():
            yield k

    def __len__(self):
        l = 0
        for i in self.iteritems():
            l += 1
        return l

    def iteritems(self):
        prefix = self._prefix
        for k in self.allowed_keys:
            try:
                yield (k, getattr(self, prefix + k))
            except AttributeError:
                pass

    def itervalues(self):
        for k, v in self.iteritems():
            yield v

    def __delitem__(self, k):
        try:
            delattr(self, self._prefix + k)
        except AttributeError:
            raise KeyError(k)

    def __setitem__(self, k, v):
        setattr(self, self._prefix + k, v)

    def setdefault(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return default

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError(
                "expected at most 1 positional argument, got " + repr(len(args))
            )
        other = None
        if args:
            other = args[0]
        if other is None:
            pass
        elif hasattr(other, "iteritems"):
            # Use getattr to avoid interference from 2to3.
            for k, v in getattr(other, "iteritems")():
                self[k] = v
        elif hasattr(other, "items"):
            # Use getattr to avoid interference from 2to3.
            for k, v in getattr(other, "items")():
                self[k] = v
        elif hasattr(other, "keys"):
            for k in other.keys():
                self[k] = other[k]
        else:
            for k, v in other:
                self[k] = v
        if kwargs:
            self.update(kwargs)

    def __getitem__(self, k):
        try:
            return getattr(self, self._prefix + k)
        except AttributeError:
            raise KeyError(k)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, k):
        return hasattr(self, self._prefix + k)

    def pop(self, key, *args):
        if len(args) > 1:
            raise TypeError(
                "pop expected at most 2 arguments, got " + repr(1 + len(args))
            )
        try:
            value = self[key]
        except KeyError:
            if args:
                return args[0]
            raise
        del self[key]
        return value

    def popitem(self):
        try:
            k, v = self.iteritems().next()
        except StopIteration:
            raise KeyError("container is empty")
        del self[k]
        return (k, v)

    def copy(self):
        c = self.__class__()
        c.update(self)
        return c

    def clear(self):
        for k in self.allowed_keys:
            try:
                delattr(self, self._prefix + k)
            except AttributeError:
                pass

    def __str__(self):
        return str(dict(self.iteritems()))

    def __repr__(self):
        return repr(dict(self.iteritems()))

    items = iteritems
    keys = __iter__
    values = itervalues


class _PickledSlotDict(_SlotDict):
    """
    Since LocalSlotDict instances are not directly picklable, this
    class exists as a way to express pickled LocalSlotDict instances,
    using a plain __dict__ instead of custom __slots__.
    """

    def __init__(self, prefix, allowed_keys, *args, **kwargs):
        self._prefix = prefix
        self.allowed_keys = allowed_keys
        super().__init__(*args, **kwargs)


_slot_dict_classes = weakref.WeakValueDictionary()


def slot_dict_class(keys, prefix="_val_"):
    """
    Generates mapping classes that behave similar to a dict but store values
    as object attributes that are allocated via __slots__. Instances of these
    objects have a smaller memory footprint than a normal dict object.

    @param keys: Fixed set of allowed keys
    @type keys: Iterable
    @param prefix: a prefix to use when mapping
            attribute names from keys
    @type prefix: String
    @rtype: SlotDict
    @return: A class that constructs SlotDict instances
            having the specified keys.
    """
    if isinstance(keys, frozenset):
        keys_set = keys
    else:
        keys_set = frozenset(keys)
    cache_key = (keys_set, prefix)
    v = _slot_dict_classes.get(cache_key)
    if v is None:

        class LocalSlotDict(_SlotDict):
            allowed_keys = keys_set
            _prefix = prefix
            __slots__ = tuple(prefix + k for k in allowed_keys)

        v = LocalSlotDict
        _slot_dict_classes[cache_key] = v
    return v



# Copyright 2010-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import re
import portage
from portage import os
from portage.const import PORTAGE_PYM_PATH
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop

from _emerge.PipeReader import PipeReader
from _emerge.SpawnProcess import SpawnProcess


class LazyImportPortageBaselineTestCase(TestCase):
    _module_re = re.compile(r"^(portage|_emerge)\.")

    _baseline_imports = frozenset(
        [
            "portage.const",
            "portage.installation",
            "portage.localization",
            "portage.proxy",
            "portage.proxy.lazyimport",
            "portage.proxy.objectproxy",
            "portage._selinux",
        ]
    )

    _baseline_import_cmd = [
        portage._python_interpreter,
        "-c",
        """
import os
import sys
sys.path.insert(0, os.environ["PORTAGE_PYM_PATH"])
import portage
sys.stdout.write(" ".join(k for k in sys.modules
	if sys.modules[k] is not None))
""",
    ]

    def testLazyImportPortageBaseline(self):
        """
        Check what modules are imported by a baseline module import.
        """

        env = os.environ.copy()
        pythonpath = env.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is None:
            pythonpath = ""
        else:
            pythonpath = ":" + pythonpath
        pythonpath = PORTAGE_PYM_PATH + pythonpath
        env["PYTHONPATH"] = pythonpath

        # If python is patched to insert the path of the
        # currently installed portage module into sys.path,
        # then the above PYTHONPATH override doesn't help.
        env["PORTAGE_PYM_PATH"] = PORTAGE_PYM_PATH

        scheduler = global_event_loop()
        master_fd, slave_fd = os.pipe()
        master_file = os.fdopen(master_fd, "rb", 0)
        slave_file = os.fdopen(slave_fd, "wb")
        producer = SpawnProcess(
            args=self._baseline_import_cmd,
            env=env,
            fd_pipes={1: slave_fd},
            scheduler=scheduler,
        )
        producer.start()
        slave_file.close()

        consumer = PipeReader(
            input_files={"producer": master_file}, scheduler=scheduler
        )

        consumer.start()
        consumer.wait()
        self.assertEqual(producer.wait(), os.EX_OK)
        self.assertEqual(consumer.wait(), os.EX_OK)

        output = consumer.getvalue().decode("ascii", "replace").split()

        unexpected_modules = " ".join(
            sorted(
                x
                for x in output
                if self._module_re.match(x) is not None
                and x not in self._baseline_imports
            )
        )

        self.assertEqual("", unexpected_modules)



# test_PackageMaskFile.py -- Portage Unit Testing Functionality
# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.env.config import PackageMaskFile
from portage.tests import TestCase, test_cps
from tempfile import mkstemp


class PackageMaskFileTestCase(TestCase):
    def testPackageMaskFile(self):
        self.BuildFile()
        try:
            f = PackageMaskFile(self.fname)
            f.load()
            for atom in f:
                self.assertTrue(atom in test_cps)
        finally:
            self.NukeFile()

    def BuildFile(self):
        fd, self.fname = mkstemp()
        f = os.fdopen(fd, "w")
        f.write("\n".join(test_cps))
        f.close()

    def NukeFile(self):
        os.unlink(self.fname)



# Copyright 2009-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["lazyimport"]

import sys
import types
import threading

from portage.proxy.objectproxy import ObjectProxy


_module_proxies = {}
_module_proxies_lock = threading.RLock()


def _register_module_proxy(name, proxy):
    _module_proxies_lock.acquire()
    try:
        proxy_list = _module_proxies.get(name)
        if proxy_list is None:
            proxy_list = []
            _module_proxies[name] = proxy_list
        proxy_list.append(proxy)
    finally:
        _module_proxies_lock.release()


def _unregister_module_proxy(name):
    """
    Destroy all proxies that reference the give module name. Also, check
    for other proxies referenced by modules that have been imported and
    destroy those proxies too. This way, destruction of a single proxy
    can trigger destruction of all the rest. If a target module appears
    to be partially imported (indicated when an AttributeError is caught),
    this function will leave in place proxies that reference it.
    """
    _module_proxies_lock.acquire()
    try:
        if name in _module_proxies:
            modules = sys.modules
            for name, proxy_list in list(_module_proxies.items()):
                if name not in modules:
                    continue
                # First delete this name from the dict so that
                # if this same thread reenters below, it won't
                # enter this path again.
                del _module_proxies[name]
                try:
                    while proxy_list:
                        proxy = proxy_list.pop()
                        object.__getattribute__(proxy, "_get_target")()
                except AttributeError:
                    # Apparently the target module is only partially
                    # imported, so proxies that reference it cannot
                    # be destroyed yet.
                    proxy_list.append(proxy)
                    _module_proxies[name] = proxy_list
    finally:
        _module_proxies_lock.release()


class _LazyImport(ObjectProxy):
    __slots__ = ("_scope", "_alias", "_name", "_target")

    def __init__(self, scope, alias, name):
        ObjectProxy.__init__(self)
        object.__setattr__(self, "_scope", scope)
        object.__setattr__(self, "_alias", alias)
        object.__setattr__(self, "_name", name)
        _register_module_proxy(name, self)

    def _get_target(self):
        try:
            return object.__getattribute__(self, "_target")
        except AttributeError:
            pass
        name = object.__getattribute__(self, "_name")
        __import__(name)
        target = sys.modules[name]
        object.__setattr__(self, "_target", target)
        object.__getattribute__(self, "_scope")[
            object.__getattribute__(self, "_alias")
        ] = target
        _unregister_module_proxy(name)
        return target


class _LazyImportFrom(_LazyImport):
    __slots__ = ("_attr_name",)

    def __init__(self, scope, name, attr_name, alias):
        object.__setattr__(self, "_attr_name", attr_name)
        _LazyImport.__init__(self, scope, alias, name)

    def _get_target(self):
        try:
            return object.__getattribute__(self, "_target")
        except AttributeError:
            pass
        name = object.__getattribute__(self, "_name")
        attr_name = object.__getattribute__(self, "_attr_name")
        __import__(name)
        try:
            target = getattr(sys.modules[name], attr_name)
        except AttributeError:
            # Try to import it as a submodule
            try:
                __import__(f"{name}.{attr_name}")
            except ImportError:
                pass
            # If it's a submodule, this will succeed. Otherwise, it may
            # be that the module is only partially imported, so raise
            # AttributeError for _unregister_module_proxy() to handle.
            target = getattr(sys.modules[name], attr_name)

        object.__setattr__(self, "_target", target)
        object.__getattribute__(self, "_scope")[
            object.__getattribute__(self, "_alias")
        ] = target
        _unregister_module_proxy(name)
        return target


def lazyimport(scope, *args):
    """
    Create a proxy in the given scope in order to performa a lazy import.

    Syntax         Result
    foo            import foo
    foo:bar,baz    from foo import bar, baz
    foo:bar@baz    from foo import bar as baz

    @param scope: the scope in which to place the import, typically globals()
    @type myfilename: dict
    @param args: module names to import
    @type args: strings
    """

    modules = sys.modules

    for s in args:
        parts = s.split(":", 1)
        if len(parts) == 1:
            name = s

            if not name or not isinstance(name, str):
                raise ValueError(name)

            components = name.split(".")
            parent_scope = scope
            for i in range(len(components)):
                alias = components[i]
                if i < len(components) - 1:
                    parent_name = ".".join(components[: i + 1])
                    __import__(parent_name)
                    mod = modules.get(parent_name)
                    if not isinstance(mod, types.ModuleType):
                        # raise an exception
                        __import__(name)
                    parent_scope[alias] = mod
                    parent_scope = mod.__dict__
                    continue

                already_imported = modules.get(name)
                if already_imported is not None:
                    parent_scope[alias] = already_imported
                else:
                    parent_scope[alias] = _LazyImport(parent_scope, alias, name)

        else:
            name, fromlist = parts
            already_imported = modules.get(name)
            fromlist = fromlist.split(",")
            for s in fromlist:
                if not s:
                    # This happens if there's an extra comma in fromlist.
                    raise ValueError("Empty module attribute name")
                alias = s.split("@", 1)
                if len(alias) == 1:
                    alias = alias[0]
                    attr_name = alias
                else:
                    attr_name, alias = alias
                if already_imported is not None:
                    try:
                        scope[alias] = getattr(already_imported, attr_name)
                    except AttributeError:
                        # Apparently the target module is only partially
                        # imported, so create a proxy.
                        already_imported = None
                        scope[alias] = _LazyImportFrom(scope, name, attr_name, alias)
                else:
                    scope[alias] = _LazyImportFrom(scope, name, attr_name, alias)



# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import platform
import tempfile

from pathlib import Path

import portage.process

from portage import shutil
from portage.tests import TestCase


class SpawnWarnLargeEnvTestCase(TestCase):
    def testSpawnWarnLargeEnv(self):
        if platform.system() != "Linux":
            self.skipTest("not Linux")

        env = dict()
        env["LARGE_ENV_VAR"] = "X" * 1024 * 96

        tmpdir = tempfile.mkdtemp()
        previous_env_too_large_warnings = portage.process.env_too_large_warnings
        try:
            logfile = tmpdir / Path("logfile")
            echo_output = "This is an echo process with a large env"
            retval = portage.process.spawn(
                ["echo", echo_output],
                env=env,
                logfile=logfile,
                warn_on_large_env=True,
            )

            with open(logfile) as f:
                logfile_content = f.read()
                self.assertIn(
                    echo_output,
                    logfile_content,
                )
            self.assertTrue(
                portage.process.env_too_large_warnings > previous_env_too_large_warnings
            )
            self.assertEqual(retval, 0)
        finally:
            shutil.rmtree(tmpdir)



# Copyright 1999-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import functools

from _emerge.AsynchronousLock import AsynchronousLock

import portage
from portage import os
from portage.exception import PortageException
from portage.util.SlotObject import SlotObject


class EbuildBuildDir(SlotObject):
    __slots__ = ("scheduler", "settings", "locked", "_catdir", "_lock_obj")

    def __init__(self, **kwargs):
        SlotObject.__init__(self, **kwargs)
        self.locked = False

    def _assert_lock(self, async_lock):
        if async_lock.returncode != os.EX_OK:
            # TODO: create a better way to propagate this error to the caller
            raise AssertionError(
                f"AsynchronousLock failed with returncode {async_lock.returncode}"
            )

    def clean_log(self):
        """Discard existing log. The log will not be be discarded
        in cases when it would not make sense, like when FEATURES=keepwork
        is enabled."""
        settings = self.settings
        if "keepwork" in settings.features:
            return
        log_file = settings.get("PORTAGE_LOG_FILE")
        if log_file is not None and os.path.isfile(log_file):
            try:
                os.unlink(log_file)
            except OSError:
                pass

    def async_lock(self):
        """
        Acquire the lock asynchronously. Notification is available
        via the add_done_callback method of the returned Future instance.

        This raises an AlreadyLocked exception if async_lock() is called
        while a lock is already held. In order to avoid this, call
        async_unlock() or check whether the "locked" attribute is True
        or False before calling async_lock().

        @returns: Future, result is None
        """
        if self._lock_obj is not None:
            raise self.AlreadyLocked((self._lock_obj,))

        dir_path = self.settings.get("PORTAGE_BUILDDIR")
        if not dir_path:
            raise AssertionError("PORTAGE_BUILDDIR is unset")
        catdir = os.path.dirname(dir_path)
        self._catdir = catdir
        catdir_lock = AsynchronousLock(path=catdir, scheduler=self.scheduler)
        builddir_lock = AsynchronousLock(path=dir_path, scheduler=self.scheduler)
        result = self.scheduler.create_future()

        def catdir_locked(catdir_lock):
            try:
                self._assert_lock(catdir_lock)
            except AssertionError as e:
                result.set_exception(e)
                return

            try:
                portage.util.ensure_dirs(
                    catdir, gid=portage.portage_gid, mode=0o70, mask=0
                )
            except PortageException as e:
                if not os.path.isdir(catdir):
                    result.set_exception(e)
                    return

            builddir_lock.addExitListener(builddir_locked)
            builddir_lock.start()

        def builddir_locked(builddir_lock):
            try:
                self._assert_lock(builddir_lock)
            except AssertionError as e:
                catdir_lock.async_unlock.add_done_callback(
                    functools.partial(catdir_unlocked, exception=e)
                )
                return

            self._lock_obj = builddir_lock
            self.locked = True
            self.settings["PORTAGE_BUILDDIR_LOCKED"] = "1"
            catdir_lock.async_unlock().add_done_callback(catdir_unlocked)

        def catdir_unlocked(future, exception=None):
            if not (exception is None and future.exception() is None):
                result.set_exception(exception or future.exception())
            else:
                result.set_result(None)

        try:
            portage.util.ensure_dirs(
                os.path.dirname(catdir), gid=portage.portage_gid, mode=0o70, mask=0
            )
        except PortageException:
            if not os.path.isdir(os.path.dirname(catdir)):
                raise

        catdir_lock.addExitListener(catdir_locked)
        catdir_lock.start()
        return result

    def async_unlock(self):
        """
        Release the lock asynchronously. Release notification is available
        via the add_done_callback method of the returned Future instance.

        @returns: Future, result is None
        """
        result = self.scheduler.create_future()

        def builddir_unlocked(future):
            if future.exception() is not None:
                result.set_exception(future.exception())
            else:
                self._lock_obj = None
                self.locked = False
                self.settings.pop("PORTAGE_BUILDDIR_LOCKED", None)
                catdir_lock = AsynchronousLock(
                    path=self._catdir, scheduler=self.scheduler
                )
                catdir_lock.addExitListener(catdir_locked)
                catdir_lock.start()

        def catdir_locked(catdir_lock):
            if catdir_lock.wait() != os.EX_OK:
                result.set_result(None)
            else:
                try:
                    os.rmdir(self._catdir)
                except OSError:
                    pass
                catdir_lock.async_unlock().add_done_callback(catdir_unlocked)

        def catdir_unlocked(future):
            if future.exception() is None:
                result.set_result(None)
            else:
                result.set_exception(future.exception())

        if self._lock_obj is None:
            self.scheduler.call_soon(result.set_result, None)
        else:
            self._lock_obj.async_unlock().add_done_callback(builddir_unlocked)
        return result

    class AlreadyLocked(portage.exception.PortageException):
        pass



# Copyright 2014-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os

from collections import OrderedDict

from portage.module import Modules
from portage.sync.controller import SyncManager
from portage.sync.config_checks import check_type

_SUBMODULE_PATH_MAP = OrderedDict(
    [
        ("glsa", ("metadata/glsa",)),
        ("news", ("metadata/news",)),
        ("profiles", ("metadata/layout.conf", "profiles")),
    ]
)

path = os.path.join(os.path.dirname(__file__), "modules")
# initial development debug info
# print("module path:", path)

module_controller = Modules(path=path, namepath="portage.sync.modules")

# initial development debug info
# print(module_controller.module_names)
module_names = module_controller.module_names[:]


def module_specific_options(repo):
    """Get the authorized module specific options set for
    the repos.conf settings for the repo"""
    global module_controller

    if repo.sync_type:
        try:
            return frozenset(
                module_controller.modules[repo.sync_type]["module_specific_options"]
            )
        except KeyError:
            pass
    return frozenset()


def validate_config(repo, logger):
    """Validate the repos.conf settings for the repo"""
    global module_names, module_controller
    if not check_type(repo, logger, module_names):
        return False

    # print(repo)
    if repo.sync_type:
        validated = module_controller.modules[repo.sync_type]["validate_config"]
        return validated(repo, logger).repo_checks()
    return True



# Copyright 2012-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import shlex
import shutil
import subprocess
import sys
import textwrap

import portage
from portage import os
from portage.const import BASH_BINARY, PORTAGE_PYM_PATH, USER_CONFIG_PATH
from portage.tests import TestCase, CommandStep, FunctionStep
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs


class PortdbCacheTestCase(TestCase):
    def testPortdbCache(self):
        debug = False

        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/A-2": {},
            "sys-apps/B-1": {},
            "sys-apps/B-2": {},
            "sys-apps/C-1": {
                "EAPI": 8,
                "MISC_CONTENT": "inherit bar foo baz",
            },
        }

        # The convoluted structure here is to test accumulation
        # of IDEPEND across eclasses (bug #870295).
        eclasses = {
            "foo": ("inherit bar",),
            "bar": ("IDEPEND=dev-libs/A",),
            "baz": ("IDEPEND=",),
        }

        playground = ResolverPlayground(ebuilds=ebuilds, eclasses=eclasses, debug=debug)
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        test_repo_location = settings.repositories["test_repo"].location
        user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)
        metadata_dir = os.path.join(test_repo_location, "metadata")
        md5_cache_dir = os.path.join(metadata_dir, "md5-cache")
        pms_cache_dir = os.path.join(metadata_dir, "cache")
        layout_conf_path = os.path.join(metadata_dir, "layout.conf")

        portage_python = portage._python_interpreter
        egencache_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.bindir), "egencache"),
            "--update-manifests",
            "--sign-manifests=n",
            "--repo",
            "test_repo",
            "--repositories-configuration",
            settings.repositories.config_string(),
        )
        python_cmd = (portage_python, "-b", "-Wd", "-c")

        test_commands = (
            FunctionStep(function=lambda _: shutil.rmtree(md5_cache_dir) or True),
            FunctionStep(
                function=lambda i: self.assertFalse(
                    os.path.exists(pms_cache_dir), f"step {i}"
                )
            ),
            FunctionStep(
                function=lambda i: self.assertFalse(
                    os.path.exists(md5_cache_dir), f"step {i}"
                )
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=python_cmd
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					if portage.portdb.repositories['test_repo'].location in portage.portdb._pregen_auxdb:
						sys.exit(1)
			        """
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=egencache_cmd + ("--update",),
            ),
            FunctionStep(
                lambda i: self.assertFalse(os.path.exists(pms_cache_dir), f"step {i}")
            ),
            FunctionStep(
                lambda i: self.assertTrue(os.path.exists(md5_cache_dir), f"step {i}")
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=python_cmd
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					if portage.portdb.repositories['test_repo'].location not in portage.portdb._pregen_auxdb:
						sys.exit(1)
			        """
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=python_cmd
                + (
                    textwrap.dedent(
                        """
						import os, sys, portage
						from portage.cache.flat_hash import md5_database
						if not isinstance(portage.portdb._pregen_auxdb[portage.portdb.repositories['test_repo'].location], md5_database):
							sys.exit(1)
			        """
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=(BASH_BINARY,)
                + (
                    "-c",
                    "echo %s > %s"
                    % tuple(
                        map(
                            shlex.quote,
                            (
                                "cache-formats = md5-dict pms",
                                layout_conf_path,
                            ),
                        )
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=egencache_cmd + ("--update",),
            ),
            FunctionStep(
                function=lambda i: self.assertTrue(
                    os.path.exists(md5_cache_dir), f"step {i}"
                )
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=python_cmd
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					if portage.portdb.repositories['test_repo'].location not in portage.portdb._pregen_auxdb:
						sys.exit(1)
			        """
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=python_cmd
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					from portage.cache.flat_hash import md5_database
					if not isinstance(portage.portdb._pregen_auxdb[portage.portdb.repositories['test_repo'].location], md5_database):
						sys.exit(1)
			        """
                    ),
                ),
            ),
            # Disable DeprecationWarnings, since the pms format triggers them
            # in portdbapi._create_pregen_cache().
            CommandStep(
                returncode=os.EX_OK,
                command=(BASH_BINARY,)
                + (
                    "-c",
                    "echo %s > %s"
                    % tuple(
                        map(
                            shlex.quote,
                            (
                                "cache-formats = pms md5-dict",
                                layout_conf_path,
                            ),
                        )
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=(portage_python,)
                + ("-b", "-Wd", "-Wi::DeprecationWarning", "-c")
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					if portage.portdb.repositories['test_repo'].location not in portage.portdb._pregen_auxdb:
						sys.exit(1)
			        """
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=(portage_python,)
                + ("-b", "-Wd", "-Wi::DeprecationWarning", "-c")
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					from portage.cache.metadata import database as pms_database
					if not isinstance(portage.portdb._pregen_auxdb[portage.portdb.repositories['test_repo'].location], pms_database):
						sys.exit(1)
			        """
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=(portage_python,)
                + ("-b", "-Wd", "-Wi::DeprecationWarning", "-c")
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					location = portage.portdb.repositories['test_repo'].location
					if not portage.portdb._pregen_auxdb[location]["sys-apps/C-1"]['IDEPEND']:
						sys.exit(1)
					"""
                    ),
                ),
            ),
            # Test auto-detection and preference for md5-cache when both
            # cache formats are available but layout.conf is absent.
            CommandStep(
                returncode=os.EX_OK,
                command=(BASH_BINARY,) + ("-c", f"rm {shlex.quote(layout_conf_path)}"),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=python_cmd
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					if portage.portdb.repositories['test_repo'].location not in portage.portdb._pregen_auxdb:
						sys.exit(1)
			        """
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=python_cmd
                + (
                    textwrap.dedent(
                        """
					import os, sys, portage
					from portage.cache.flat_hash import md5_database
					if not isinstance(portage.portdb._pregen_auxdb[portage.portdb.repositories['test_repo'].location], md5_database):
						sys.exit(1)
			        """
                    ),
                ),
            ),
        )

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PATH": settings["PATH"],
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
        }

        if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
            env["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
                "__PORTAGE_TEST_HARDLINK_LOCKS"
            ]

        dirs = [user_config_dir]

        try:
            for d in dirs:
                ensure_dirs(d)

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, step in enumerate(test_commands):
                if isinstance(step, FunctionStep):
                    try:
                        step.function(i)
                    except Exception as e:
                        if isinstance(e, AssertionError) and f"step {i}" in str(e):
                            raise
                        raise AssertionError(
                            f"step {i} raised {e.__class__.__name__}"
                        ) from e
                    continue

                proc = subprocess.Popen(
                    step.command,
                    env=dict(env.items(), **(step.env or {})),
                    cwd=step.cwd,
                    stdout=stdout,
                )

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != step.returncode:
                        for line in output:
                            sys.stderr.write(portage._unicode_decode(line))

                self.assertEqual(
                    step.returncode,
                    proc.returncode,
                    f"{step.command} (step {i}) failed with exit code {proc.returncode}",
                )
        finally:
            playground.cleanup()



# elog/mod_echo.py - elog dispatch module
# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys
from portage.output import EOutput, colorize
from portage.const import EBUILD_PHASES
from portage.localization import _


_items = []


def process(mysettings, key, logentries, fulltext):
    global _items
    logfile = None
    # output logfile explicitly only if it isn't in tempdir, otherwise
    # it will be removed anyway
    if (
        key == mysettings.mycpv
        and "PORTAGE_LOGDIR" in mysettings
        and "PORTAGE_LOG_FILE" in mysettings
    ):
        logfile = mysettings["PORTAGE_LOG_FILE"]

    try:
        binary = mysettings.configdict["pkg"]["MERGE_TYPE"] == "binary"
    except KeyError:
        binary = False
    _items.append((mysettings["ROOT"], key, logentries, logfile, binary))


def finalize():
    # For consistency, send all message types to stdout.
    sys.stdout.flush()
    sys.stderr.flush()
    stderr = sys.stderr
    try:
        sys.stderr = sys.stdout
        _finalize()
    finally:
        sys.stderr = stderr
        sys.stdout.flush()
        sys.stderr.flush()


def _finalize():
    global _items
    printer = EOutput()
    for root, key, logentries, logfile, binary in _items:
        color = "PKG_BINARY_MERGE" if binary else "INFORM"

        print()

        if root == "/":
            printer.einfo(_("Messages for package %s:") % colorize(color, key))
        else:
            printer.einfo(
                _("Messages for package %(pkg)s merged to %(root)s:")
                % {"pkg": colorize(color, key), "root": root}
            )
        if logfile is not None:
            printer.einfo(_("Log file: %s") % colorize("INFORM", logfile))
        print()
        for phase in EBUILD_PHASES:
            if phase not in logentries:
                continue
            for msgtype, msgcontent in logentries[phase]:
                fmap = {
                    "INFO": printer.einfo,
                    "LOG": printer.elog,
                    "WARN": printer.ewarn,
                    "ERROR": printer.eerror,
                    "QA": printer.eqawarn,
                }
                if isinstance(msgcontent, str):
                    msgcontent = [msgcontent]
                for line in msgcontent:
                    fmap[msgtype](line.strip("\n"))
    _items = []



# Copyright 2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os
from portage.const import GLOBAL_CONFIG_PATH

COMPAT_FEATURES = 'FEATURES="${FEATURES} -binpkg-multi-instance"'


def main():
    """
    If the current installation is still has binpkg-multi-instance
    disabled, then patch make.globals inside ${ED} to maintain backward
    compatibility. This is intended to be called from the ebuild as
    follows:

    pkg_preinst() {
            python_setup
            env -u FEATURES -u PORTAGE_REPOSITORIES \
                    PYTHONPATH="${D}$(python_get_sitedir)${PYTHONPATH:+:${PYTHONPATH}}" \
                    "${PYTHON}" -m portage._compat_upgrade.binpkg_multi_instance || die
    }
    """
    if "binpkg-multi-instance" not in portage.settings.features:
        portage.output.EOutput().einfo(
            f"Setting make.globals default {COMPAT_FEATURES} for backward compatibility"
        )
        config_path = os.path.join(
            os.environ["ED"], GLOBAL_CONFIG_PATH.lstrip(os.sep), "make.globals"
        )
        with open(config_path, "a") as f:
            f.write(f"{COMPAT_FEATURES}\n")


if __name__ == "__main__":
    main()



# Copyright 1999-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.AbstractDepPriority import AbstractDepPriority


class UnmergeDepPriority(AbstractDepPriority):
    __slots__ = (
        "cross",
        "ignored",
        "optional",
        "satisfied",
    )
    """
    Combination of properties           Priority  Category

    installtime                            0       HARD
    runtime_slot_op                       -1       HARD
    runtime                               -2       HARD
    runtime_post                          -3       HARD
    buildtime                             -4       SOFT
    (none of the above)                   -4       SOFT
    """

    MAX = 0
    SOFT = -4
    MIN = -4

    def __init__(self, **kwargs):
        AbstractDepPriority.__init__(self, **kwargs)
        if self.buildtime:
            self.optional = True

    def __int__(self):
        if self.installtime:
            return 0
        if self.runtime_slot_op:
            return -1
        if self.runtime:
            return -2
        if self.runtime_post:
            return -3
        if self.buildtime:
            return -4
        return -4

    def __str__(self):
        if self.ignored:
            return "ignored"
        if self.installtime:
            return "install time"
        if self.runtime_slot_op:
            return "hard slot op"
        myvalue = self.__int__()
        if myvalue > self.SOFT:
            return "hard"
        return "soft"



# Copyright 2010-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import bz2
import fnmatch
import subprocess
import tempfile
import portage

from itertools import permutations
from portage import os
from portage import shutil
from portage.const import (
    GLOBAL_CONFIG_PATH,
    PORTAGE_BIN_PATH,
    USER_CONFIG_PATH,
    SUPPORTED_GENTOO_BINPKG_FORMATS,
)
from portage.process import find_binary
from portage.dep import Atom, _repo_separator
from portage.dbapi.bintree import binarytree
from portage._sets import load_default_config
from portage._sets.base import InternalPackageSet
from portage.tests import cnf_path
from portage.util import ensure_dirs, normalize_path
from portage.versions import catsplit
from portage.exception import InvalidBinaryPackageFormat
from portage.gpg import GPG

import _emerge
from _emerge.actions import _calc_depclean, expand_set_arguments
from _emerge.Blocker import Blocker
from _emerge.create_depgraph_params import create_depgraph_params
from _emerge.DependencyArg import DependencyArg
from _emerge.depgraph import (
    _frozen_depgraph_config,
    backtrack_depgraph,
)
from _emerge.Package import Package
from _emerge.RootConfig import RootConfig


class ResolverPlayground:
    """
    This class helps to create the necessary files on disk and
    the needed settings instances, etc. for the resolver to do
    its work.
    """

    config_files = frozenset(
        (
            "eapi",
            "binrepos.conf",
            "layout.conf",
            "make.conf",
            "modules",
            "package.accept_keywords",
            "package.env",
            "package.keywords",
            "package.license",
            "package.mask",
            "package.properties",
            "package.provided",
            "packages",
            "package.unmask",
            "package.use",
            "package.use.force",
            "package.use.mask",
            "package.use.stable",
            "package.use.stable.force",
            "package.use.stable.mask",
            "soname.provided",
            "use.force",
            "use.mask",
            "use.stable",
            "layout.conf",
        )
    )

    metadata_xml_template = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE pkgmetadata SYSTEM "https://www.gentoo.org/dtd/metadata.dtd">
<pkgmetadata>
<maintainer type="person">
<email>maintainer-needed@gentoo.org</email>
<description>Description of the maintainership</description>
</maintainer>
<longdescription>Long description of the package</longdescription>
<use>
%(flags)s
</use>
</pkgmetadata>
"""

    portage_bin = (
        "ebuild",
        "egencache",
        "emerge",
        "emerge-webrsync",
        "emirrordist",
        "glsa-check",
        "portageq",
        "quickpkg",
    )

    portage_sbin = (
        "archive-conf",
        "dispatch-conf",
        "emaint",
        "env-update",
        "etc-update",
        "fixpackages",
        "regenworld",
    )

    def __init__(
        self,
        ebuilds={},
        binpkgs={},
        installed={},
        profile={},
        repo_configs={},
        user_config={},
        sets={},
        world=[],
        world_sets=[],
        distfiles={},
        eclasses={},
        eprefix=None,
        targetroot=False,
        debug=False,
    ):
        """
        ebuilds: cpv -> metadata mapping simulating available ebuilds.
        installed: cpv -> metadata mapping simulating installed packages.
                If a metadata key is missing, it gets a default value.
        profile: settings defined by the profile.
        """

        self.debug = debug
        if True:
            if eprefix is None:
                self.eprefix = normalize_path(tempfile.mkdtemp())
            else:
                self.eprefix = normalize_path(eprefix)

            # EPREFIX/bin is used by fake true_binaries. Real binaries goes into EPREFIX/usr/bin
            eubin = os.path.join(self.eprefix, "usr", "bin")
            ensure_dirs(eubin)
            os.symlink(portage._python_interpreter, os.path.join(eubin, "python"))
            for x in self.portage_bin:
                os.symlink(os.path.join(PORTAGE_BIN_PATH, x), os.path.join(eubin, x))

            eusbin = os.path.join(self.eprefix, "usr", "sbin")
            ensure_dirs(eusbin)
            for x in self.portage_sbin:
                os.symlink(os.path.join(PORTAGE_BIN_PATH, x), os.path.join(eusbin, x))

            essential_binaries = (
                "awk",
                "basename",
                "bash",
                "bzip2",
                "cat",
                "chgrp",
                "chmod",
                "chown",
                "comm",
                "cp",
                "egrep",
                "env",
                "find",
                "flock",
                "grep",
                "head",
                "install",
                "ln",
                "mkdir",
                "mkfifo",
                "mktemp",
                "mv",
                "readlink",
                "rm",
                "sed",
                "sort",
                "gtar",
                "tr",
                "uname",
                "uniq",
                "xargs",
                "zstd",
            )
            # Exclude internal wrappers from PATH lookup.
            orig_path = os.environ["PATH"]
            included_paths = []
            for path in orig_path.split(":"):
                if path and not fnmatch.fnmatch(path, "*/portage/*/ebuild-helpers*"):
                    included_paths.append(path)
            try:
                os.environ["PATH"] = ":".join(included_paths)
                for x in essential_binaries:
                    path = find_binary(x)
                    if path is None:
                        raise portage.exception.CommandNotFound(x)
                    os.symlink(path, os.path.join(eubin, x))
            finally:
                os.environ["PATH"] = orig_path

        # Tests may override portage.const.EPREFIX in order to
        # simulate a prefix installation. It's reasonable to do
        # this because tests should be self-contained such that
        # the "real" value of portage.const.EPREFIX is entirely
        # irrelevant (see bug #492932).
        self._orig_eprefix = portage.const.EPREFIX
        portage.const.EPREFIX = self.eprefix.rstrip(os.sep)

        self.eroot = self.eprefix + os.sep
        if targetroot:
            self.target_root = os.path.join(self.eroot, "target_root")
        else:
            self.target_root = os.sep
        self.distdir = os.path.join(self.eroot, "var", "portage", "distfiles")
        self.pkgdir = os.path.join(self.eprefix, "pkgdir")
        self.vdbdir = os.path.join(self.eroot, "var/db/pkg")
        os.makedirs(self.vdbdir)

        if not debug:
            portage.util.noiselimit = -2

        self._repositories = {}
        # Make sure the main repo is always created
        self._get_repo_dir("test_repo")

        self._create_distfiles(distfiles)
        self._create_ebuilds(ebuilds)
        self._create_installed(installed)
        self._create_profile(
            ebuilds, eclasses, installed, profile, repo_configs, user_config, sets
        )
        self._create_world(world, world_sets)

        self.settings, self.trees = self._load_config()

        self.gpg = None
        self._create_binpkgs(binpkgs)
        self._create_ebuild_manifests(ebuilds)

        portage.util.noiselimit = 0

    def reload_config(self):
        """
        Reload configuration from disk, which is useful if it has
        been modified after the constructor has been called.
        """
        for eroot in self.trees:
            portdb = self.trees[eroot]["porttree"].dbapi
            portdb.close_caches()
        self.settings, self.trees = self._load_config()

    def _get_repo_dir(self, repo):
        """
        Create the repo directory if needed.
        """
        if repo not in self._repositories:
            if repo == "test_repo":
                self._repositories["DEFAULT"] = {"main-repo": repo}

            repo_path = os.path.join(self.eroot, "var", "repositories", repo)
            self._repositories[repo] = {"location": repo_path}
            profile_path = os.path.join(repo_path, "profiles")

            try:
                os.makedirs(profile_path)
            except OSError:
                pass

            repo_name_file = os.path.join(profile_path, "repo_name")
            with open(repo_name_file, "w") as f:
                f.write(f"{repo}\n")

        return self._repositories[repo]["location"]

    def _create_distfiles(self, distfiles):
        os.makedirs(self.distdir)
        for k, v in distfiles.items():
            with open(os.path.join(self.distdir, k), "wb") as f:
                f.write(v)

    def _create_ebuilds(self, ebuilds):
        for cpv in ebuilds:
            a = Atom("=" + cpv, allow_repo=True)
            repo = a.repo
            if repo is None:
                repo = "test_repo"

            metadata = ebuilds[cpv].copy()
            copyright_header = metadata.pop("COPYRIGHT_HEADER", None)
            eapi = metadata.pop("EAPI", "0")
            misc_content = metadata.pop("MISC_CONTENT", None)
            metadata.setdefault("DEPEND", "")
            metadata.setdefault("SLOT", "0")
            metadata.setdefault("KEYWORDS", "x86")
            metadata.setdefault("IUSE", "")

            unknown_keys = set(metadata).difference(portage.dbapi.dbapi._known_keys)
            if unknown_keys:
                raise ValueError(
                    "metadata of ebuild '%s' contains unknown keys: %s"
                    % (cpv, sorted(unknown_keys))
                )

            repo_dir = self._get_repo_dir(repo)
            ebuild_dir = os.path.join(repo_dir, a.cp)
            ebuild_path = os.path.join(ebuild_dir, a.cpv.split("/")[1] + ".ebuild")
            try:
                os.makedirs(ebuild_dir)
            except OSError:
                pass

            with open(ebuild_path, "w") as f:
                if copyright_header is not None:
                    f.write(copyright_header)
                f.write(f'EAPI="{eapi}"\n')
                for k, v in metadata.items():
                    f.write(f'{k}="{v}"\n')
                if misc_content is not None:
                    f.write(misc_content)

    def _create_ebuild_manifests(self, ebuilds):
        for repo_name in self._repositories:
            if repo_name == "DEFAULT":
                continue
            egencache_cmd = [
                "egencache",
                f"--repo={repo_name}",
                "--update",
                "--update-manifests",
                "--sign-manifests=n",
                "--strict-manifests=n",
                f"--repositories-configuration={self.settings['PORTAGE_REPOSITORIES']}",
                f"--jobs={portage.util.cpuinfo.get_cpu_count()}",
            ]
            result = subprocess.run(
                egencache_cmd,
                env=self.settings.environ(),
            )
            if result.returncode != os.EX_OK:
                raise AssertionError(
                    f"command failed with returncode {result.returncode}: {egencache_cmd}"
                )

    def _create_binpkgs(self, binpkgs):
        # When using BUILD_ID, there can be multiple instances for the
        # same cpv. Therefore, binpkgs may be an iterable instead of
        # a dict.
        items = getattr(binpkgs, "items", None)
        items = items() if items is not None else binpkgs
        binpkg_format = self.settings.get(
            "BINPKG_FORMAT", SUPPORTED_GENTOO_BINPKG_FORMATS[0]
        )
        if binpkg_format == "gpkg":
            if self.gpg is None:
                self.gpg = GPG(self.settings)
                self.gpg.unlock()
        for cpv, metadata in items:
            a = Atom("=" + cpv, allow_repo=True)
            repo = a.repo
            if repo is None:
                repo = "test_repo"

            pn = catsplit(a.cp)[1]
            cat, pf = catsplit(a.cpv)
            metadata = metadata.copy()
            metadata.setdefault("SLOT", "0")
            metadata.setdefault("KEYWORDS", "x86")
            metadata.setdefault("BUILD_TIME", "0")
            metadata["repository"] = repo
            metadata["CATEGORY"] = cat
            metadata["PF"] = pf
            metadata["BINPKG_FORMAT"] = binpkg_format

            repo_dir = self.pkgdir
            category_dir = os.path.join(repo_dir, cat)
            if "BUILD_ID" in metadata:
                if binpkg_format == "xpak":
                    binpkg_path = os.path.join(
                        category_dir, pn, f"{pf}-{metadata['BUILD_ID']}.xpak"
                    )
                elif binpkg_format == "gpkg":
                    binpkg_path = os.path.join(
                        category_dir,
                        pn,
                        f"{pf}-{metadata['BUILD_ID']}.gpkg.tar",
                    )
                else:
                    raise InvalidBinaryPackageFormat(binpkg_format)
            else:
                if binpkg_format == "xpak":
                    binpkg_path = os.path.join(category_dir, pf + ".tbz2")
                elif binpkg_format == "gpkg":
                    binpkg_path = os.path.join(category_dir, pf + ".gpkg.tar")
                else:
                    raise InvalidBinaryPackageFormat(binpkg_format)

            ensure_dirs(os.path.dirname(binpkg_path))
            if binpkg_format == "xpak":
                t = portage.xpak.tbz2(binpkg_path)
                t.recompose_mem(portage.xpak.xpak_mem(metadata))
            elif binpkg_format == "gpkg":
                t = portage.gpkg.gpkg(self.settings, a.cpv, binpkg_path)
                t.compress(os.path.dirname(binpkg_path), metadata)
            else:
                raise InvalidBinaryPackageFormat(binpkg_format)

            bintree = binarytree(pkgdir=self.pkgdir, settings=self.settings)
            bintree.populate(force_reindex=True)

    def _create_installed(self, installed):
        for cpv in installed:
            a = Atom("=" + cpv, allow_repo=True)
            repo = a.repo
            if repo is None:
                repo = "test_repo"

            vdb_pkg_dir = os.path.join(self.vdbdir, a.cpv)
            try:
                os.makedirs(vdb_pkg_dir)
            except OSError:
                pass

            metadata = installed[cpv].copy()
            metadata.setdefault("SLOT", "0")
            metadata.setdefault("BUILD_TIME", "0")
            metadata.setdefault("COUNTER", "0")
            metadata.setdefault("KEYWORDS", "~x86")

            unknown_keys = set(metadata).difference(portage.dbapi.dbapi._known_keys)
            unknown_keys.discard("BUILD_TIME")
            unknown_keys.discard("BUILD_ID")
            unknown_keys.discard("COUNTER")
            unknown_keys.discard("repository")
            unknown_keys.discard("USE")
            unknown_keys.discard("PROVIDES")
            unknown_keys.discard("REQUIRES")
            if unknown_keys:
                raise ValueError(
                    "metadata of installed '%s' contains unknown keys: %s"
                    % (cpv, sorted(unknown_keys))
                )

            metadata["repository"] = repo
            for k, v in metadata.items():
                with open(os.path.join(vdb_pkg_dir, k), "w") as f:
                    f.write(f"{v}\n")

            ebuild_path = os.path.join(vdb_pkg_dir, a.cpv.split("/")[1] + ".ebuild")
            with open(ebuild_path, "w") as f:
                f.write(f"EAPI=\"{metadata.pop('EAPI', '0')}\"\n")
                for k, v in metadata.items():
                    f.write(f'{k}="{v}"\n')

            env_path = os.path.join(vdb_pkg_dir, "environment.bz2")
            with bz2.BZ2File(env_path, mode="w") as f:
                with open(ebuild_path, "rb") as inputfile:
                    f.write(inputfile.read())

    def _create_profile(
        self, ebuilds, eclasses, installed, profile, repo_configs, user_config, sets
    ):
        user_config_dir = os.path.join(self.eroot, USER_CONFIG_PATH)

        try:
            os.makedirs(user_config_dir)
        except OSError:
            pass

        for repo in self._repositories:
            if repo == "DEFAULT":
                continue

            repo_dir = self._get_repo_dir(repo)
            profile_dir = os.path.join(repo_dir, "profiles")
            metadata_dir = os.path.join(repo_dir, "metadata")
            os.makedirs(metadata_dir)

            # Create $REPO/profiles/categories
            categories = set()
            for cpv in ebuilds:
                ebuilds_repo = Atom("=" + cpv, allow_repo=True).repo
                if ebuilds_repo is None:
                    ebuilds_repo = "test_repo"
                if ebuilds_repo == repo:
                    categories.add(catsplit(cpv)[0])

            categories_file = os.path.join(profile_dir, "categories")
            with open(categories_file, "w") as f:
                for cat in categories:
                    f.write(cat + "\n")

            # Create $REPO/profiles/license_groups
            license_file = os.path.join(profile_dir, "license_groups")
            with open(license_file, "w") as f:
                f.write("EULA TEST\n")

            repo_config = repo_configs.get(repo)
            if repo_config:
                for config_file, lines in repo_config.items():
                    if config_file not in self.config_files and not any(
                        fnmatch.fnmatch(config_file, os.path.join(x, "*"))
                        for x in self.config_files
                    ):
                        raise ValueError(f"Unknown config file: '{config_file}'")

                    if config_file in ("layout.conf",):
                        file_name = os.path.join(repo_dir, "metadata", config_file)
                    else:
                        file_name = os.path.join(profile_dir, config_file)
                        if "/" in config_file and not os.path.isdir(
                            os.path.dirname(file_name)
                        ):
                            os.makedirs(os.path.dirname(file_name))
                    with open(file_name, "w") as f:
                        for line in lines:
                            f.write(f"{line}\n")
                        # Temporarily write empty value of masters until it becomes default.
                        # TODO: Delete all references to "# use implicit masters" when empty value becomes default.
                        if config_file == "layout.conf" and not any(
                            line.startswith(("masters =", "# use implicit masters"))
                            for line in lines
                        ):
                            f.write("masters =\n")

            # Create $profile_dir/eclass (we fail to digest the ebuilds if it's not there)
            eclass_dir = os.path.join(repo_dir, "eclass")
            os.makedirs(eclass_dir)

            for eclass_name, eclass_content in eclasses.items():
                with open(os.path.join(eclass_dir, f"{eclass_name}.eclass"), "w") as f:
                    if isinstance(eclass_content, str):
                        eclass_content = [eclass_content]
                    for line in eclass_content:
                        f.write(f"{line}\n")

            # Temporarily write empty value of masters until it becomes default.
            if not repo_config or "layout.conf" not in repo_config:
                layout_conf_path = os.path.join(repo_dir, "metadata", "layout.conf")
                with open(layout_conf_path, "w") as f:
                    f.write("masters =\n")

            if repo == "test_repo":
                # Create a minimal profile in /var/db/repos/gentoo
                sub_profile_dir = os.path.join(
                    profile_dir, "default", "linux", "x86", "test_profile"
                )
                os.makedirs(sub_profile_dir)

                if not (profile and "eapi" in profile):
                    eapi_file = os.path.join(sub_profile_dir, "eapi")
                    with open(eapi_file, "w") as f:
                        f.write("0\n")

                make_defaults_file = os.path.join(sub_profile_dir, "make.defaults")
                with open(make_defaults_file, "w") as f:
                    f.write('ARCH="x86"\n')
                    f.write('ACCEPT_KEYWORDS="x86"\n')

                use_force_file = os.path.join(sub_profile_dir, "use.force")
                with open(use_force_file, "w") as f:
                    f.write("x86\n")

                parent_file = os.path.join(sub_profile_dir, "parent")
                with open(parent_file, "w") as f:
                    f.write("..\n")

                if profile:
                    for config_file, lines in profile.items():
                        if config_file not in self.config_files:
                            raise ValueError(f"Unknown config file: '{config_file}'")

                        file_name = os.path.join(sub_profile_dir, config_file)
                        with open(file_name, "w") as f:
                            for line in lines:
                                f.write(f"{line}\n")

                # Create profile symlink
                os.symlink(
                    sub_profile_dir, os.path.join(user_config_dir, "make.profile")
                )

        gpg_test_path = os.environ["PORTAGE_GNUPGHOME"]

        make_conf = {
            "ACCEPT_KEYWORDS": "x86",
            "BINPKG_GPG_SIGNING_BASE_COMMAND": f"flock {gpg_test_path}/portage-binpkg-gpg.lock /usr/bin/gpg --sign --armor --yes --pinentry-mode loopback --passphrase GentooTest [PORTAGE_CONFIG]",
            "BINPKG_GPG_SIGNING_GPG_HOME": gpg_test_path,
            "BINPKG_GPG_SIGNING_KEY": "0x5D90EA06352177F6",
            "BINPKG_GPG_VERIFY_GPG_HOME": gpg_test_path,
            "CLEAN_DELAY": "0",
            "DISTDIR": self.distdir,
            "EMERGE_WARNING_DELAY": "0",
            "FEATURES": "${FEATURES} binpkg-signing gpg-keepalive",
            "PKGDIR": self.pkgdir,
            "PORTAGE_INST_GID": str(portage.data.portage_gid),
            "PORTAGE_INST_UID": str(portage.data.portage_uid),
            "PORTAGE_TMPDIR": os.path.join(self.eroot, "var/tmp"),
        }

        # Pass along PORTAGE_USERNAME and PORTAGE_GRPNAME since they
        # need to be inherited by ebuild subprocesses.
        if "PORTAGE_USERNAME" in os.environ:
            make_conf["PORTAGE_USERNAME"] = os.environ["PORTAGE_USERNAME"]
        if "PORTAGE_GRPNAME" in os.environ:
            make_conf["PORTAGE_GRPNAME"] = os.environ["PORTAGE_GRPNAME"]

        make_conf_lines = []
        for k_v in make_conf.items():
            make_conf_lines.append('%s="%s"' % k_v)

        if "make.conf" in user_config:
            make_conf_lines.extend(user_config["make.conf"])
            if "BINPKG_FORMAT=gpkg" in user_config["make.conf"]:
                make_conf_lines.append(
                    'FEATURES="${FEATURES} binpkg-request-signature"'
                )

        if not portage.process.sandbox_capable or os.environ.get("SANDBOX_ON") == "1":
            # avoid problems from nested sandbox instances
            make_conf_lines.append('FEATURES="${FEATURES} -sandbox -usersandbox"')

        configs = user_config.copy()
        configs["make.conf"] = make_conf_lines

        for config_file, lines in configs.items():
            if config_file not in self.config_files:
                raise ValueError(f"Unknown config file: '{config_file}'")

            file_name = os.path.join(user_config_dir, config_file)
            with open(file_name, "w") as f:
                for line in lines:
                    f.write(f"{line}\n")

        # Create /usr/share/portage/config/make.globals
        make_globals_path = os.path.join(
            self.eroot, GLOBAL_CONFIG_PATH.lstrip(os.sep), "make.globals"
        )
        ensure_dirs(os.path.dirname(make_globals_path))
        os.symlink(os.path.join(str(cnf_path), "make.globals"), make_globals_path)

        # Create /usr/share/portage/config/sets/portage.conf
        default_sets_conf_dir = os.path.join(
            self.eroot, "usr/share/portage/config/sets"
        )

        try:
            os.makedirs(default_sets_conf_dir)
        except OSError:
            pass

        provided_sets_portage_conf = os.path.join(str(cnf_path), "sets", "portage.conf")
        os.symlink(
            provided_sets_portage_conf,
            os.path.join(default_sets_conf_dir, "portage.conf"),
        )

        set_config_dir = os.path.join(user_config_dir, "sets")

        try:
            os.makedirs(set_config_dir)
        except OSError:
            pass

        for sets_file, lines in sets.items():
            file_name = os.path.join(set_config_dir, sets_file)
            with open(file_name, "w") as f:
                for line in lines:
                    f.write(f"{line}\n")

    def _create_world(self, world, world_sets):
        # Create /var/lib/portage/world
        var_lib_portage = os.path.join(self.eroot, "var", "lib", "portage")
        os.makedirs(var_lib_portage)

        world_file = os.path.join(var_lib_portage, "world")
        world_set_file = os.path.join(var_lib_portage, "world_sets")

        with open(world_file, "w") as f:
            for atom in world:
                f.write(f"{atom}\n")

        with open(world_set_file, "w") as f:
            for atom in world_sets:
                f.write(f"{atom}\n")

    def _load_config(self):
        create_trees_kwargs = {}
        if self.target_root != os.sep:
            create_trees_kwargs["target_root"] = self.target_root

        env = {
            "PATH": f"{self.eprefix}/usr/sbin:{self.eprefix}/usr/bin:{os.environ['PATH']}",
            "PORTAGE_REPOSITORIES": "\n".join(
                "[%s]\n%s"
                % (
                    repo_name,
                    "\n".join(f"{k} = {v}" for k, v in repo_config.items()),
                )
                for repo_name, repo_config in self._repositories.items()
            ),
        }

        if self.debug:
            env["PORTAGE_DEBUG"] = "1"

        trees = portage.create_trees(
            env=env, eprefix=self.eprefix, **create_trees_kwargs
        )

        for root, root_trees in trees.items():
            settings = root_trees["vartree"].settings
            settings._init_dirs()
            setconfig = load_default_config(settings, root_trees)
            root_trees["root_config"] = RootConfig(settings, root_trees, setconfig)

        return trees[trees._target_eroot]["vartree"].settings, trees

    def run(self, atoms, options={}, action=None):
        options = options.copy()
        options["--pretend"] = True
        if self.debug:
            options["--debug"] = True

        if action is None:
            if options.get("--depclean"):
                action = "depclean"
            elif options.get("--prune"):
                action = "prune"

        if "--usepkgonly" in options:
            options["--usepkg"] = True

        global_noiselimit = portage.util.noiselimit
        global_emergelog_disable = _emerge.emergelog._disable
        try:
            if not self.debug:
                portage.util.noiselimit = -2
            _emerge.emergelog._disable = True

            # NOTE: frozen_config could be cached and reused if options and params were constant.
            params_action = (
                "remove" if action in ("dep_check", "depclean", "prune") else action
            )
            params = create_depgraph_params(options, params_action)
            frozen_config = _frozen_depgraph_config(
                self.settings, self.trees, options, params, None
            )

            atoms, retval = expand_set_arguments(
                atoms, action, self.trees[self.eroot]["root_config"]
            )
            if retval != os.EX_OK:
                raise AssertionError(
                    f"expand_set_arguments failed with retval {retval}"
                )

            if params_action == "remove":
                depclean_result = _calc_depclean(
                    self.settings,
                    self.trees,
                    None,
                    options,
                    action,
                    InternalPackageSet(initial_atoms=atoms, allow_wildcard=True),
                    None,
                    frozen_config=frozen_config,
                )
                result = ResolverPlaygroundDepcleanResult(
                    atoms,
                    depclean_result.returncode,
                    depclean_result.cleanlist,
                    depclean_result.ordered,
                    depclean_result.req_pkg_count,
                    depclean_result.depgraph,
                )
            else:
                success, depgraph, favorites = backtrack_depgraph(
                    self.settings,
                    self.trees,
                    options,
                    params,
                    action,
                    atoms,
                    None,
                    frozen_config=frozen_config,
                )
                depgraph._show_merge_list()
                depgraph.display_problems()
                result = ResolverPlaygroundResult(atoms, success, depgraph, favorites)
        finally:
            portage.util.noiselimit = global_noiselimit
            _emerge.emergelog._disable = global_emergelog_disable

        return result

    def run_TestCase(self, test_case):
        if not isinstance(test_case, ResolverPlaygroundTestCase):
            raise TypeError("ResolverPlayground needs a ResolverPlaygroundTestCase")
        for atoms in test_case.requests:
            result = self.run(atoms, test_case.options, test_case.action)
            if not test_case.compare_with_result(result):
                return

    def cleanup(self):
        if self.gpg is not None:
            self.gpg.stop()
        for eroot in self.trees:
            portdb = self.trees[eroot]["porttree"].dbapi
            portdb.close_caches()
        if self.debug:
            print(f"\nEROOT={self.eroot}")
        else:
            shutil.rmtree(self.eroot)
        if hasattr(self, "_orig_eprefix"):
            portage.const.EPREFIX = self._orig_eprefix


class ResolverPlaygroundTestCase:
    def __init__(self, request, **kwargs):
        self.all_permutations = kwargs.pop("all_permutations", False)
        self.ignore_mergelist_order = kwargs.pop("ignore_mergelist_order", False)
        self.ignore_cleanlist_order = kwargs.pop("ignore_cleanlist_order", False)
        self.ambiguous_merge_order = kwargs.pop("ambiguous_merge_order", False)
        self.ambiguous_slot_collision_solutions = kwargs.pop(
            "ambiguous_slot_collision_solutions", False
        )
        self.check_repo_names = kwargs.pop("check_repo_names", False)
        self.merge_order_assertions = kwargs.pop("merge_order_assertions", False)

        if self.all_permutations:
            self.requests = list(permutations(request))
        else:
            self.requests = [request]

        self.options = kwargs.pop("options", {})
        self.action = kwargs.pop("action", None)
        self.test_success = True
        self.fail_msg = None
        self._checks = kwargs.copy()

    def compare_with_result(self, result):
        checks = dict.fromkeys(result.checks)
        for key, value in self._checks.items():
            if not key in checks:
                raise KeyError(f"Not an available check: '{key}'")
            checks[key] = value

        fail_msgs = []
        for key, value in checks.items():
            got = getattr(result, key)
            expected = value

            if key in result.optional_checks and expected is None:
                continue

            if key == "mergelist":
                if not self.check_repo_names:
                    # Strip repo names if we don't check them
                    if got:
                        new_got = []
                        for cpv in got:
                            if cpv[:1] == "!":
                                new_got.append(cpv)
                                continue
                            new_got.append(cpv.split(_repo_separator)[0])
                        got = new_got
                    if expected:
                        new_expected = []
                        for obj in expected:
                            if isinstance(obj, str):
                                if obj[:1] == "!":
                                    new_expected.append(obj)
                                    continue
                                new_expected.append(obj.split(_repo_separator)[0])
                                continue
                            new_expected.append(set())
                            for cpv in obj:
                                if cpv[:1] != "!":
                                    cpv = cpv.split(_repo_separator)[0]
                                new_expected[-1].add(cpv)
                        expected = new_expected
                if self.ignore_mergelist_order and got is not None:
                    got = set(got)
                    expected = set(expected)

                if self.ambiguous_merge_order and got:
                    expected_stack = list(reversed(expected))
                    got_stack = list(reversed(got))
                    new_expected = []
                    match = True
                    while got_stack and expected_stack:
                        got_token = got_stack.pop()
                        expected_obj = expected_stack.pop()
                        if isinstance(expected_obj, str):
                            new_expected.append(expected_obj)
                            if got_token == expected_obj:
                                continue
                            # result doesn't match, so stop early
                            match = False
                            break
                        expected_obj = set(expected_obj)
                        try:
                            expected_obj.remove(got_token)
                        except KeyError:
                            # result doesn't match, so stop early
                            match = False
                            break
                        new_expected.append(got_token)
                        while got_stack and expected_obj:
                            got_token = got_stack.pop()
                            try:
                                expected_obj.remove(got_token)
                            except KeyError:
                                match = False
                                break
                            new_expected.append(got_token)
                        if not match:
                            # result doesn't match, so stop early
                            break
                        if expected_obj:
                            # result does not match, so stop early
                            match = False
                            new_expected.append(tuple(expected_obj))
                            break
                    if expected_stack:
                        # result does not match, add leftovers to new_expected
                        match = False
                        expected_stack.reverse()
                        new_expected.extend(expected_stack)
                    expected = new_expected

                    if match and self.merge_order_assertions:
                        for node1, node2 in self.merge_order_assertions:
                            if not got.index(node1) < got.index(node2):
                                fail_msgs.append(
                                    "atoms: ("
                                    + ", ".join(result.atoms)
                                    + "), key: "
                                    + (
                                        "merge_order_assertions, expected: %s"
                                        % str((node1, node2))
                                    )
                                    + ", got: "
                                    + str(got)
                                )

            elif key == "cleanlist" and self.ignore_cleanlist_order:
                got = set(got)
                expected = set(expected)

            elif (
                key == "slot_collision_solutions"
                and self.ambiguous_slot_collision_solutions
            ):
                # Tests that use all_permutations can have multiple
                # outcomes here.
                for x in expected:
                    if x == got:
                        expected = x
                        break
            elif (
                key
                in (
                    "unstable_keywords",
                    "needed_p_mask_changes",
                    "unsatisfied_deps",
                    "required_use_unsatisfied",
                )
                and expected is not None
            ):
                # unsatisfied_deps can be a dict for depclean-like actions
                expected = expected if isinstance(expected, dict) else set(expected)

            elif key == "forced_rebuilds" and expected is not None:
                expected = {k: set(v) for k, v in expected.items()}

            if got != expected:
                fail_msgs.append(
                    "atoms: ("
                    + ", ".join(result.atoms)
                    + "), key: "
                    + key
                    + ", expected: "
                    + str(expected)
                    + ", got: "
                    + str(got)
                )
        if fail_msgs:
            self.test_success = False
            self.fail_msg = "\n".join(fail_msgs)
            return False
        return True


def _mergelist_str(x, depgraph):
    if isinstance(x, DependencyArg):
        mergelist_str = x.arg
    elif isinstance(x, Blocker):
        mergelist_str = x.atom
    else:
        repo_str = ""
        if x.repo != "test_repo":
            repo_str = _repo_separator + x.repo
        build_id_str = ""
        if x.type_name == "binary" and x.cpv.build_id is not None:
            build_id_str = f"-{x.cpv.build_id}"
        mergelist_str = x.cpv + build_id_str + repo_str
        if x.built:
            if x.operation == "merge":
                desc = x.type_name
            else:
                desc = x.operation
            mergelist_str = f"[{desc}]{mergelist_str}"
        if x.root != depgraph._frozen_config._running_root.root:
            mergelist_str += "{targetroot}"
    return mergelist_str


class ResolverPlaygroundResult:
    checks = (
        "success",
        "mergelist",
        "use_changes",
        "license_changes",
        "unstable_keywords",
        "slot_collision_solutions",
        "circular_dependency_solutions",
        "needed_p_mask_changes",
        "unsatisfied_deps",
        "forced_rebuilds",
        "required_use_unsatisfied",
        "graph_order",
        "virtual_cycle",
    )
    optional_checks = (
        "forced_rebuilds",
        "required_use_unsatisfied",
        "unsatisfied_deps",
        "graph_order",
    )

    def __init__(self, atoms, success, mydepgraph, favorites):
        self.atoms = atoms
        self.success = success
        self.depgraph = mydepgraph
        self.favorites = favorites
        self.mergelist = None
        self.use_changes = None
        self.license_changes = None
        self.unstable_keywords = None
        self.needed_p_mask_changes = None
        self.slot_collision_solutions = None
        self.circular_dependency_solutions = None
        self.unsatisfied_deps = frozenset()
        self.forced_rebuilds = None
        self.required_use_unsatisfied = None
        self.virtual_cycle = None

        self.graph_order = [
            _mergelist_str(node, self.depgraph)
            for node in self.depgraph._dynamic_config.digraph
        ]

        if self.depgraph._dynamic_config._serialized_tasks_cache is not None:
            self.mergelist = []
            for x in self.depgraph._dynamic_config._serialized_tasks_cache:
                self.mergelist.append(_mergelist_str(x, self.depgraph))

        if self.depgraph._dynamic_config._needed_use_config_changes:
            self.use_changes = {}
            for (
                pkg,
                needed_use_config_changes,
            ) in self.depgraph._dynamic_config._needed_use_config_changes.items():
                new_use, changes = needed_use_config_changes
                self.use_changes[pkg.cpv] = changes

        if self.depgraph._dynamic_config._needed_unstable_keywords:
            self.unstable_keywords = set()
            for pkg in self.depgraph._dynamic_config._needed_unstable_keywords:
                self.unstable_keywords.add(pkg.cpv)

        if self.depgraph._dynamic_config._needed_p_mask_changes:
            self.needed_p_mask_changes = set()
            for pkg in self.depgraph._dynamic_config._needed_p_mask_changes:
                self.needed_p_mask_changes.add(pkg.cpv)

        if self.depgraph._dynamic_config._needed_license_changes:
            self.license_changes = {}
            for (
                pkg,
                missing_licenses,
            ) in self.depgraph._dynamic_config._needed_license_changes.items():
                self.license_changes[pkg.cpv] = missing_licenses

        if self.depgraph._dynamic_config._slot_conflict_handler is not None:
            self.slot_collision_solutions = []
            handler = self.depgraph._dynamic_config._slot_conflict_handler

            for change in handler.changes:
                new_change = {}
                for pkg in change:
                    new_change[pkg.cpv] = change[pkg]
                self.slot_collision_solutions.append(new_change)

        if self.depgraph._dynamic_config._circular_dependency_handler is not None:
            handler = self.depgraph._dynamic_config._circular_dependency_handler
            sol = handler.solutions
            self.circular_dependency_solutions = dict(
                zip([x.cpv for x in sol.keys()], sol.values())
            )

        if self.depgraph._dynamic_config._unsatisfied_deps_for_display:
            self.unsatisfied_deps = {
                dep_info[0][1]
                for dep_info in self.depgraph._dynamic_config._unsatisfied_deps_for_display
            }

        if self.depgraph._forced_rebuilds:
            self.forced_rebuilds = {
                child.cpv: {parent.cpv for parent in parents}
                for child_dict in self.depgraph._forced_rebuilds.values()
                for child, parents in child_dict.items()
            }

        required_use_unsatisfied = []
        for (
            pargs,
            kwargs,
        ) in self.depgraph._dynamic_config._unsatisfied_deps_for_display:
            if "show_req_use" in kwargs:
                required_use_unsatisfied.append(pargs[1])
        if required_use_unsatisfied:
            self.required_use_unsatisfied = set(required_use_unsatisfied)

        if self.depgraph._virtual_cycle:
            self.virtual_cycle = {pkg.cpv for pkg in self.depgraph._virtual_cycle}


class ResolverPlaygroundDepcleanResult:
    checks = (
        "success",
        "cleanlist",
        "ordered",
        "req_pkg_count",
        "graph_order",
        "unsatisfied_deps",
    )
    optional_checks = (
        "cleanlist",
        "ordered",
        "req_pkg_count",
        "graph_order",
        "unsatisfied_deps",
    )

    def __init__(self, atoms, rval, cleanlist, ordered, req_pkg_count, depgraph):
        self.atoms = atoms
        self.success = rval == 0
        self.cleanlist = cleanlist
        self.ordered = ordered
        self.req_pkg_count = req_pkg_count
        self.graph_order = [
            _mergelist_str(node, depgraph) for node in depgraph._dynamic_config.digraph
        ]
        self.unsatisfied_deps = {}
        for dep in depgraph._dynamic_config._initially_unsatisfied_deps:
            if isinstance(dep.parent, Package):
                parent_repr = dep.parent.cpv
            else:
                parent_repr = dep.parent.arg
            self.unsatisfied_deps.setdefault(parent_repr, set()).add(dep.atom)



# Copyright 2014-2015, 2023 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from functools import partial
import shlex
import shutil
import stat
import subprocess
import sys
import time

import portage
from portage import os
from portage import _encodings, _unicode_decode
from portage.const import BASH_BINARY, PORTAGE_PYM_PATH
from portage.process import find_binary
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs, find_updated_config_files


class ConfigProtectTestCase(TestCase):
    def testConfigProtect(self):
        """
        Demonstrates many different scenarios. For example:

         * regular file replaces regular file
         * regular file replaces symlink
         * regular file replaces directory
         * symlink replaces symlink
         * symlink replaces regular file
         * symlink replaces directory
         * directory replaces regular file
         * directory replaces symlink
        """

        debug = False

        content_A_1 = """
S="${WORKDIR}"

src_install() {
	insinto /etc/A
	keepdir /etc/A/dir_a
	keepdir /etc/A/symlink_replaces_dir
	keepdir /etc/A/regular_replaces_dir
	echo regular_a_1 > "${T}"/regular_a
	doins "${T}"/regular_a
	echo regular_b_1 > "${T}"/regular_b
	doins "${T}"/regular_b
	dosym regular_a /etc/A/regular_replaces_symlink
	dosym regular_b /etc/A/symlink_replaces_symlink
	echo regular_replaces_regular_1 > \
		"${T}"/regular_replaces_regular
	doins "${T}"/regular_replaces_regular
	echo symlink_replaces_regular > \
		"${T}"/symlink_replaces_regular
	doins "${T}"/symlink_replaces_regular
}

"""

        content_A_2 = """
S="${WORKDIR}"

src_install() {
	insinto /etc/A
	keepdir /etc/A/dir_a
	dosym dir_a /etc/A/symlink_replaces_dir
	echo regular_replaces_dir > "${T}"/regular_replaces_dir
	doins "${T}"/regular_replaces_dir
	echo regular_a_2 > "${T}"/regular_a
	doins "${T}"/regular_a
	echo regular_b_2 > "${T}"/regular_b
	doins "${T}"/regular_b
	echo regular_replaces_symlink > \
		"${T}"/regular_replaces_symlink
	doins "${T}"/regular_replaces_symlink
	dosym regular_b /etc/A/symlink_replaces_symlink
	echo regular_replaces_regular_2 > \
		"${T}"/regular_replaces_regular
	doins "${T}"/regular_replaces_regular
	dosym regular_a /etc/A/symlink_replaces_regular
}

"""

        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "5",
                "IUSE": "+flag",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "MISC_CONTENT": content_A_1,
            },
            "dev-libs/A-2": {
                "EAPI": "5",
                "IUSE": "+flag",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "MISC_CONTENT": content_A_2,
            },
        }

        playground = ResolverPlayground(ebuilds=ebuilds, debug=debug)
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")

        portage_python = portage._python_interpreter
        dispatch_conf_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.sbindir), "dispatch-conf"),
        )
        emerge_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.bindir), "emerge"),
        )
        etc_update_cmd = (BASH_BINARY, os.path.join(str(self.sbindir), "etc-update"))
        etc_update_auto = etc_update_cmd + (
            "--automode",
            "-5",
        )

        config_protect = "/etc"

        def modify_files(dir_path):
            for name in os.listdir(dir_path):
                path = os.path.join(dir_path, name)
                st = os.lstat(path)
                if stat.S_ISREG(st.st_mode):
                    with open(path, mode="a", encoding=_encodings["stdio"]) as f:
                        f.write("modified at %d\n" % time.time())
                elif stat.S_ISLNK(st.st_mode):
                    old_dest = os.readlink(path)
                    os.unlink(path)
                    os.symlink(old_dest + " modified at %d" % time.time(), path)

        def updated_config_files(count):
            self.assertEqual(
                count,
                sum(
                    len(x[1])
                    for x in find_updated_config_files(
                        eroot, shlex.split(config_protect)
                    )
                ),
            )

        test_commands = (
            etc_update_cmd,
            dispatch_conf_cmd,
            emerge_cmd + ("-1", "=dev-libs/A-1"),
            partial(updated_config_files, 0),
            emerge_cmd + ("-1", "=dev-libs/A-2"),
            partial(updated_config_files, 2),
            etc_update_auto,
            partial(updated_config_files, 0),
            emerge_cmd + ("-1", "=dev-libs/A-2"),
            partial(updated_config_files, 0),
            # Test bug #523684, where a file renamed or removed by the
            # admin forces replacement files to be merged with config
            # protection.
            partial(shutil.rmtree, os.path.join(eprefix, "etc", "A")),
            emerge_cmd + ("-1", "=dev-libs/A-2"),
            partial(updated_config_files, 8),
            etc_update_auto,
            partial(updated_config_files, 0),
            # Modify some config files, and verify that it triggers
            # config protection.
            partial(modify_files, os.path.join(eroot, "etc", "A")),
            emerge_cmd + ("-1", "=dev-libs/A-2"),
            partial(updated_config_files, 6),
            etc_update_auto,
            partial(updated_config_files, 0),
            # Modify some config files, downgrade to A-1, and verify
            # that config protection works properly when the file
            # types are changing.
            partial(modify_files, os.path.join(eroot, "etc", "A")),
            emerge_cmd + ("-1", "--noconfmem", "=dev-libs/A-1"),
            partial(updated_config_files, 6),
            etc_update_auto,
            partial(updated_config_files, 0),
        )

        distdir = playground.distdir
        fake_bin = os.path.join(eprefix, "bin")
        portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")

        path = settings.get("PATH")
        if path is not None and not path.strip():
            path = None
        if path is None:
            path = ""
        else:
            path = ":" + path
        path = fake_bin + path

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "CLEAN_DELAY": "0",
            "CONFIG_PROTECT": config_protect,
            "DISTDIR": distdir,
            "EMERGE_DEFAULT_OPTS": "-v",
            "EMERGE_WARNING_DELAY": "0",
            "ETC_UPDATE_OS_FAMILY": "gentoo",
            "INFODIR": "",
            "INFOPATH": "",
            "PATH": path,
            "PORTAGE_INST_GID": str(os.getgid()),  # str(portage.data.portage_gid),
            "PORTAGE_INST_UID": str(os.getuid()),  # str(portage.data.portage_uid),
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PORTAGE_TMPDIR": portage_tmpdir,
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
            "__PORTAGE_TEST_PATH_OVERRIDE": fake_bin,
        }

        if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
            env["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
                "__PORTAGE_TEST_HARDLINK_LOCKS"
            ]

        dirs = [distdir, fake_bin, portage_tmpdir, var_cache_edb]
        etc_symlinks = ("dispatch-conf.conf", "etc-update.conf")
        # Override things that may be unavailable, or may have portability
        # issues when running tests in exotic environments.
        #   prepstrip - bug #447810 (bash read builtin EINTR problem)
        true_symlinks = ["prepstrip", "scanelf"]
        true_binary = find_binary("true")
        self.assertEqual(true_binary is None, False, "true command not found")
        try:
            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(true_binary, os.path.join(fake_bin, x))
            for x in etc_symlinks:
                os.symlink(
                    os.path.join(str(self.cnf_etc_path), x),
                    os.path.join(eprefix, "etc", x),
                )
            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for args in test_commands:
                if hasattr(args, "__call__"):
                    args()
                    continue

                if isinstance(args[0], dict):
                    local_env = env.copy()
                    local_env.update(args[0])
                    args = args[1:]
                else:
                    local_env = env

                proc = subprocess.Popen(args, env=local_env, stdout=stdout)

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != os.EX_OK:
                        for line in output:
                            sys.stderr.write(_unicode_decode(line))

                self.assertEqual(
                    os.EX_OK, proc.returncode, f"emerge failed with args {args}"
                )
        finally:
            playground.cleanup()



# Copyright 2005-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os, _unicode_encode
from portage.const import MERGING_IDENTIFIER, EPREFIX, PRIVATE_PATH, VDB_PATH
from portage.dep import isvalidatom

import shutil
import subprocess
import sys
import time


class TrackingFile:
    """File for keeping track of failed merges."""

    def __init__(self, tracking_path):
        """
        Create a TrackingFile object.

        @param tracking_path: file path used to keep track of failed merges
        @type tracking_path: String
        """
        self._tracking_path = (
            tracking_path if portage.utf8_mode else _unicode_encode(tracking_path)
        )

    def save(self, failed_pkgs):
        """
        Save the specified packages that failed to merge.

        @param failed_pkgs: dictionary of failed packages
        @type failed_pkgs: dict
        """
        tracking_path = self._tracking_path
        lines = [f"{pkg} {mtime}" for pkg, mtime in failed_pkgs.items()]
        portage.util.write_atomic(tracking_path, "\n".join(lines))

    def load(self):
        """
        Load previously failed merges.

        @rtype: dict
        @return: dictionary of packages that failed to merge
        """
        tracking_path = self._tracking_path
        if not self.exists():
            return {}
        failed_pkgs = {}
        with open(tracking_path) as tracking_file:
            for failed_merge in tracking_file:
                pkg, mtime = failed_merge.strip().split()
                failed_pkgs[pkg] = mtime
        return failed_pkgs

    def exists(self):
        """
        Check if tracking file exists.

        @rtype: bool
        @return: true if tracking file exists, false otherwise
        """
        return os.path.exists(self._tracking_path)

    def purge(self):
        """Delete previously saved tracking file if one exists."""
        if self.exists():
            os.remove(self._tracking_path)

    def __iter__(self):
        """
        Provide an iterator over failed merges.

        @return: iterator of packages that failed to merge
        """
        return self.load().items().__iter__()


class MergesHandler:
    """Handle failed package merges."""

    short_desc = "Remove failed merges"

    @staticmethod
    def name():
        return "merges"

    def __init__(self):
        """Create MergesHandler object."""
        eroot = portage.settings["EROOT"]
        tracking_path = os.path.join(eroot, PRIVATE_PATH, "failed-merges")
        self._tracking_file = TrackingFile(tracking_path)
        self._vardb_path = os.path.join(eroot, VDB_PATH)

    def can_progressbar(self, func):
        return func == "check"

    def _scan(self, onProgress=None):
        """
        Scan the file system for failed merges and return any found.

        @param onProgress: function to call for updating progress
        @type onProgress: Function
        @rtype: dict
        @return: dictionary of packages that failed to merges
        """
        failed_pkgs = {}
        for cat in os.listdir(self._vardb_path):
            pkgs_path = os.path.join(self._vardb_path, cat)
            if not os.path.isdir(pkgs_path):
                continue
            pkgs = os.listdir(pkgs_path)
            maxval = len(pkgs)
            for i, pkg in enumerate(pkgs):
                if onProgress:
                    onProgress(maxval, i + 1)
                if MERGING_IDENTIFIER in pkg:
                    mtime = int(os.stat(os.path.join(pkgs_path, pkg)).st_mtime)
                    pkg = os.path.join(cat, pkg)
                    failed_pkgs[pkg] = mtime
        return failed_pkgs

    def _failed_pkgs(self, onProgress=None):
        """
        Return failed packages from both the file system and tracking file.

        @rtype: dict
        @return: dictionary of packages that failed to merges
        """
        failed_pkgs = self._scan(onProgress)
        for pkg, mtime in self._tracking_file:
            if pkg not in failed_pkgs:
                failed_pkgs[pkg] = mtime
        return failed_pkgs

    def _remove_failed_dirs(self, failed_pkgs):
        """
        Remove the directories of packages that failed to merge.

        @param failed_pkgs: failed packages whose directories to remove
        @type failed_pkg: dict
        """
        for failed_pkg in failed_pkgs:
            pkg_path = os.path.join(self._vardb_path, failed_pkg)
            # delete failed merge directory if it exists (it might not exist
            # if loaded from tracking file)
            if os.path.exists(pkg_path):
                shutil.rmtree(pkg_path)
            # TODO: try removing package CONTENTS to prevent orphaned
            # files

    def _get_pkg_atoms(self, failed_pkgs, pkg_atoms, pkg_invalid_entries):
        """
        Get the package atoms for the specified failed packages.

        @param failed_pkgs: failed packages to iterate
        @type failed_pkgs: dict
        @param pkg_atoms: add package atoms to this set
        @type pkg_atoms: set
        @param pkg_invalid_entries: add any packages that are invalid to this set
        @type pkg_invalid_entries: set
        """

        portdb = portage.db[portage.root]["porttree"].dbapi
        for failed_pkg in failed_pkgs:
            # validate pkg name
            pkg_name = f"{failed_pkg.replace(MERGING_IDENTIFIER, '')}"
            pkg_atom = f"={pkg_name}"

            if not isvalidatom(pkg_atom):
                pkg_invalid_entries.add(f"'{pkg_atom}' is an invalid package atom.")
            if not portdb.cpv_exists(pkg_name):
                pkg_invalid_entries.add(
                    f"'{pkg_name}' does not exist in the ebuild repository."
                )
            pkg_atoms.add(pkg_atom)

    def _emerge_pkg_atoms(self, module_output, pkg_atoms, yes=False):
        """
        Emerge the specified packages atoms.

        @param module_output: output will be written to
        @type module_output: Class
        @param pkg_atoms: packages atoms to emerge
        @type pkg_atoms: set
        @param yes: do not prompt for emerge invocations
        @type yes: bool
        @rtype: list
        @return: List of results
        """
        # TODO: rewrite code to use portage's APIs instead of a subprocess
        env = {
            "FEATURES": "-collision-protect -protect-owned",
            "PATH": os.environ["PATH"],
        }
        emerge_cmd = (
            portage._python_interpreter,
            "-b",
            os.path.join(EPREFIX or "/", "usr", "bin", "emerge"),
            "--ask=n" if yes else "--ask",
            "--quiet",
            "--oneshot",
            "--complete-graph=y",
        )
        results = []
        msg = "Re-Emerging packages that failed to merge...\n"
        if module_output:
            module_output.write(msg)
        else:
            module_output = subprocess.PIPE
            results.append(msg)
        proc = subprocess.Popen(
            emerge_cmd + tuple(pkg_atoms),
            env=env,
            stdout=module_output,
            stderr=sys.stderr,
        )
        output = proc.communicate()[0]
        if output:
            results.append(output)
        if proc.returncode != os.EX_OK:
            emerge_status = f"Failed to emerge '{' '.join(pkg_atoms)}'"
        else:
            emerge_status = f"Successfully emerged '{' '.join(pkg_atoms)}'"
        results.append(emerge_status)
        return results

    def check(self, **kwargs):
        """Check for failed merges."""
        onProgress = kwargs.get("onProgress", None)
        failed_pkgs = self._failed_pkgs(onProgress)
        errors = []
        for pkg, mtime in failed_pkgs.items():
            mtime_str = time.ctime(int(mtime))
            errors.append(f"'{pkg}' failed to merge on '{mtime_str}'")
        if errors:
            return (False, errors)
        return (True, None)

    def fix(self, **kwargs):
        """Attempt to fix any failed merges."""
        module_output = kwargs.get("module_output", None)
        failed_pkgs = self._failed_pkgs()
        if not failed_pkgs:
            return (True, ["No failed merges found."])

        pkg_invalid_entries = set()
        pkg_atoms = set()
        self._get_pkg_atoms(failed_pkgs, pkg_atoms, pkg_invalid_entries)
        if pkg_invalid_entries:
            return (False, pkg_invalid_entries)

        try:
            self._tracking_file.save(failed_pkgs)
        except OSError as ex:
            errors = [f"Unable to save failed merges to tracking file: {str(ex)}\n"]
            errors.append(", ".join(sorted(failed_pkgs)))
            return (False, errors)
        self._remove_failed_dirs(failed_pkgs)
        results = self._emerge_pkg_atoms(
            module_output, pkg_atoms, yes=kwargs.get("options", {}).get("yes", False)
        )
        # list any new failed merges
        for pkg in sorted(self._scan()):
            results.append(f"'{pkg}' still found as a failed merge.")
        # reload config and remove successful packages from tracking file
        portage._reset_legacy_globals()
        vardb = portage.db[portage.root]["vartree"].dbapi
        still_failed_pkgs = {}
        for pkg, mtime in failed_pkgs.items():
            pkg_name = f"{pkg.replace(MERGING_IDENTIFIER, '')}"
            if not vardb.cpv_exists(pkg_name):
                still_failed_pkgs[pkg] = mtime
        self._tracking_file.save(still_failed_pkgs)
        if still_failed_pkgs:
            return (False, results)
        return (True, results)

    def purge(self, **kwargs):
        """Attempt to remove previously saved tracking file."""
        if not self._tracking_file.exists():
            return (True, ["Tracking file not found."])
        self._tracking_file.purge()
        return (True, ["Removed tracking file."])



# Copyright 2010-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import shlex
import textwrap

from portage import os
from portage import _python_interpreter
from portage.const import EBUILD_SH_BINARY
from portage.package.ebuild.config import config
from portage.package.ebuild.doebuild import spawn as doebuild_spawn
from portage.package.ebuild._spawn_nofetch import spawn_nofetch
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util._async.SchedulerInterface import SchedulerInterface
from portage.util._eventloop.global_event_loop import global_event_loop
from _emerge.EbuildPhase import EbuildPhase
from _emerge.MiscFunctionsProcess import MiscFunctionsProcess
from _emerge.Package import Package


class DoebuildSpawnTestCase(TestCase):
    """
    Invoke portage.package.ebuild.doebuild.spawn() with a
    minimal environment. This gives coverage to some of
    the ebuild execution internals, like ebuild.sh,
    AbstractEbuildProcess, and EbuildIpcDaemon.
    """

    def testDoebuildSpawn(self):
        ebuild_body = textwrap.dedent(
            """
			pkg_nofetch() { : ; }
		"""
        )

        ebuilds = {
            "sys-apps/portage-2.1": {
                "EAPI": "2",
                "IUSE": "build doc epydoc python3 selinux",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "RDEPEND": ">=app-shells/bash-3.2_p17 >=dev-lang/python-2.6",
                "SLOT": "0",
                "MISC_CONTENT": ebuild_body,
            }
        }

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            root_config = playground.trees[playground.eroot]["root_config"]
            portdb = root_config.trees["porttree"].dbapi
            settings = config(clone=playground.settings)
            if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
                settings["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
                    "__PORTAGE_TEST_HARDLINK_LOCKS"
                ]
                settings.backup_changes("__PORTAGE_TEST_HARDLINK_LOCKS")

            cpv = "sys-apps/portage-2.1"
            metadata = dict(
                zip(Package.metadata_keys, portdb.aux_get(cpv, Package.metadata_keys))
            )

            pkg = Package(
                built=False,
                cpv=cpv,
                installed=False,
                metadata=metadata,
                root_config=root_config,
                type_name="ebuild",
            )
            settings.setcpv(pkg)
            settings["PORTAGE_PYTHON"] = _python_interpreter
            settings["PORTAGE_BUILDDIR"] = os.path.join(settings["PORTAGE_TMPDIR"], cpv)
            settings["PYTHONDONTWRITEBYTECODE"] = os.environ.get(
                "PYTHONDONTWRITEBYTECODE", ""
            )
            settings["HOME"] = os.path.join(settings["PORTAGE_BUILDDIR"], "homedir")
            settings["T"] = os.path.join(settings["PORTAGE_BUILDDIR"], "temp")
            for x in ("PORTAGE_BUILDDIR", "HOME", "T"):
                os.makedirs(settings[x])
            os.makedirs(os.path.join(settings["PORTAGE_BUILDDIR"], ".ipc"))
            # Create a fake environment, to pretend as if the ebuild
            # has been sourced already.
            open(os.path.join(settings["T"], "environment"), "wb").close()

            scheduler = SchedulerInterface(global_event_loop())
            self.assertTrue(scheduler._loop is global_event_loop()._loop)
            for phase in ("_internal_test",):
                # Test EbuildSpawnProcess by calling doebuild.spawn() with
                # returnpid=False. This case is no longer used by portage
                # internals since EbuildPhase is used instead and that passes
                # returnpid=True to doebuild.spawn().
                rval = doebuild_spawn(
                    "%s %s"
                    % (
                        shlex.quote(
                            os.path.join(
                                settings["PORTAGE_BIN_PATH"],
                                os.path.basename(EBUILD_SH_BINARY),
                            )
                        ),
                        phase,
                    ),
                    settings,
                    free=1,
                )
                self.assertEqual(rval, os.EX_OK)

                ebuild_phase = EbuildPhase(
                    background=False,
                    phase=phase,
                    scheduler=scheduler,
                    settings=settings,
                )
                ebuild_phase.start()
                ebuild_phase.wait()
                self.assertEqual(ebuild_phase.returncode, os.EX_OK)

            ebuild_phase = MiscFunctionsProcess(
                background=False,
                commands=["success_hooks"],
                scheduler=scheduler,
                settings=settings,
            )
            ebuild_phase.start()
            ebuild_phase.wait()
            self.assertEqual(ebuild_phase.returncode, os.EX_OK)

            spawn_nofetch(portdb, portdb.findname(cpv), settings=settings)
        finally:
            playground.cleanup()
